"""TotalConnectClient() in this file is the primary class of this package.
Instantiate it like this:

usercodes = { 'default': '1234' }
client = TCC.TotalConnectClient(username, password, usercodes)

for location in client.locations:
    ### do stuff with this location
"""

import logging
import time
import warnings

import zeep

from .const import ArmType
from .exceptions import (
    AuthenticationError,
    BadResultCodeError,
    InvalidSessionError,
    RetryableTotalConnectError,
    TotalConnectError,
)
from .location import TotalConnectLocation
from .user import TotalConnectUser

PROJECT_URL = "https://github.com/craigjmidwinter/total-connect-client"

DEFAULT_USERCODE = "-1"

LOGGER = logging.getLogger(__name__)


class TotalConnectClient:
    """Client for Total Connect."""

    SUCCESS = 0
    ARM_SUCCESS = 4500
    DISARM_SUCCESS = 4500
    SESSION_INITIATED = 4500
    CONNECTION_ERROR = 4101
    FAILED_TO_CONNECT = -4104
    USER_CODE_INVALID = -4106
    USER_CODE_UNAVAILABLE = -4114
    COMMAND_FAILED = -4502
    INVALID_SESSION = -102
    INVALID_SESSIONID = -30002
    BAD_USER_OR_PASSWORD = -50004
    AUTHENTICATION_FAILED = -100
    FEATURE_NOT_SUPPORTED = -120

    MAX_RETRY_ATTEMPTS = 10

    def __init__(
        self,
        username,
        password,
        usercodes=None,
        auto_bypass_battery=False,
        retry_delay=3,  # seconds between retries
    ):
        """Initialize."""
        self.times = {}
        self.time_start = time.time()
        self.soap_client = None
        self.application_id = "14588"
        self.application_version = "1.0.34"

        self.username = username
        self.password = password
        self.usercodes = usercodes or {}
        self.auto_bypass_low_battery = auto_bypass_battery
        self.retry_delay = retry_delay

        self.token = None
        self._invalid_credentials = False
        self._module_flags = None
        self._user = None
        self._locations = {}
        self._locations_unfetched = {}

        self.authenticate()

        self.times["__init__"] = time.time() - self.time_start

    @property
    def locations(self):
        """Raises an exception if the panel cannot be reached to retrieve
        metadata or details. This can be retried later and will succeed
        if/when the panel becomes reachable.
        """
        # to_fetch is needed because items() is invalidated by del
        to_fetch = list(self._locations_unfetched.items())
        for (locationid, location) in to_fetch:
            try:
                location.get_partition_details()
                location.get_zone_details()
                location.get_panel_meta_data()
                # if we get here, it has been fetched successfully
                del self._locations_unfetched[locationid]
            except Exception:
                LOGGER.error(f"exception during initial fetch of {locationid}")
                raise
        assert not self._locations_unfetched
        return self._locations

    def __str__(self):
        """Return a text string that is printable."""
        data = (
            f"CLIENT\n\n"
            f"Username: {self.username}\n"
            f"Password: {'[hidden]' if self.password else '[unset]'}\n"
            f"Usercode: {self.usercodes}\n"
            f"Auto Bypass Low Battery: {self.auto_bypass_low_battery}\n"
            f"Invalid Credentials: {self._invalid_credentials}\n"
            f"Module Flags:\n"
        )

        for key, value in self._module_flags.items():
            data = data + f"  {key}: {value}\n"

        data = data + str(self._user)

        locations = f"LOCATIONS: {len(self.locations)}\n\n"
        for location in self.locations:
            locations = locations + str(self.locations[location])

        return data + locations

    def get_times(self):
        """Return a string with times."""
        self.times["total running time"] = time.time() - self.time_start
        msg = "total-connect-client time info (seconds):\n"
        for key, value in self.times.items():
            msg = msg + f"  {key}: {value}\n"

        return msg

    def _raise_for_retry(self, response):
        """Used internally to determine which responses should be retried in
        request().
        """
        rc = response["ResultCode"]
        if rc == self.INVALID_SESSION:
            raise InvalidSessionError("invalid session", response)
        if rc == self.INVALID_SESSIONID:
            raise InvalidSessionError("invalid session ID", response)
        if rc == self.CONNECTION_ERROR:
            raise RetryableTotalConnectError("connection error", response)
        if rc == self.FAILED_TO_CONNECT:
            raise RetryableTotalConnectError("failed to connect with panel", response)

    def raise_for_resultcode(self, response):
        """If response.ResultCode indicates success, return and do nothing.
        If it indicates an authentication error, raise AuthenticationError.
        """
        rc = response["ResultCode"]
        if rc in (
            self.SUCCESS,
            self.ARM_SUCCESS,
            self.DISARM_SUCCESS,
            self.SESSION_INITIATED,
        ):
            return
        self._raise_for_retry(response)
        if rc == self.COMMAND_FAILED:
            raise BadResultCodeError("command failed", response)
        if rc == self.FEATURE_NOT_SUPPORTED:
            raise BadResultCodeError("feature not supported", response)
        if rc == self.USER_CODE_INVALID:
            raise BadResultCodeError("user code invalid", response)
        if rc == self.BAD_USER_OR_PASSWORD:
            raise AuthenticationError("bad user or password", response)
        if rc == self.AUTHENTICATION_FAILED:
            raise AuthenticationError("authentication failed", response)
        if rc == self.USER_CODE_UNAVAILABLE:
            raise AuthenticationError("user code unavailable", response)
        raise BadResultCodeError(f"unknown result code {rc}", response)

    def request(self, request, attempts=0):
        """Send a SOAP request."""

        if not self.soap_client:
            self.soap_client = zeep.Client(
                "https://rs.alarmnet.com/TC21api/tc2.asmx?WSDL"
            )
        try:
            LOGGER.debug(f"sending API request {request}")
            r = eval("self.soap_client.service." + request)
            response = zeep.helpers.serialize_object(r)
            self._raise_for_retry(response)
            return response
        except RetryableTotalConnectError as err:
            if attempts > self.MAX_RETRY_ATTEMPTS:
                raise
            LOGGER.info(f"retrying {err.args[0]}, attempt # {attempts}")
            time.sleep(self.retry_delay)
            return self.request(request, attempts + 1)
        except InvalidSessionError:
            if attempts > self.MAX_RETRY_ATTEMPTS:
                raise
            LOGGER.info(f"reauthenticating session, attempt # {attempts}")
            self.token = None
            self.authenticate()
            return self.request(request, attempts + 1)

    def authenticate(self):
        """Login to the system. Upon success, self.token is a valid credential
        for further API calls, and self._user and self.locations are valid.
        self.locations will not be refreshed if it was non-empty on entry.
        """
        start_time = time.time()
        if self._invalid_credentials:
            raise AuthenticationError(
                f"not authenticating: password already failed for user {self.username}"
            )

        # LoginAndGetSessionDetails is very slow, so only use it when necessary
        verb = (
            "AuthenticateUserLogin" if self._locations else "LoginAndGetSessionDetails"
        )
        response = self.request(
            verb + "(self.username, self.password, "
            "self.application_id, self.application_version)"
        )
        try:
            self.raise_for_resultcode(response)
        except AuthenticationError:
            self._invalid_credentials = True
            self.token = None
            raise

        self.token = response["SessionID"]
        if not self._locations:
            self._module_flags = dict(
                x.split("=") for x in response["ModuleFlags"].split(",")
            )
            self._user = TotalConnectUser(response["UserInfo"])
            self._locations_unfetched = self._make_locations(response)
            self._locations = self._locations_unfetched.copy()
            if not self._locations:
                raise TotalConnectError("no locations found", response)
        LOGGER.info(f"{self.username} authenticated: {len(self._locations)} locations")
        self.times["authenticate()"] = time.time() - start_time

    def validate_usercode(self, device_id, usercode):
        """Return True if the usercode is valid for the device."""
        response = self.request(
            f"ValidateUserCode(self.token, {device_id}, '{usercode}')"
        )

        if response["ResultCode"] in (
            self.USER_CODE_INVALID,
            self.USER_CODE_UNAVAILABLE,
        ):
            LOGGER.warning(f"usercode {usercode} invalid for device {device_id}")
            return False
        self.raise_for_resultcode(response)
        return True

    def is_logged_in(self):
        """Return true if the client is logged into Total Connect service."""
        return self.token is not None

    def log_out(self):
        """Upon return, we are logged out. Raises TotalConnectError if we
        still might be logged in.
        """
        if self.is_logged_in():
            response = self.request("Logout(self.token)")
            self.raise_for_resultcode(response)
            LOGGER.info("Logout Successful")
            self.token = None

    def is_valid_credentials(self):
        """Return True if the credentials are known to be valid."""
        return self.token is not None

    def _make_locations(self, response):
        """Return a dict mapping LocationID to TotalConnectLocation."""
        start_time = time.time()
        new_locations = {}

        for location in (response["Locations"] or {}).get("LocationInfoBasic", {}):
            location_id = location["LocationID"]
            new_locations[location_id] = TotalConnectLocation(location, self)

            # set auto_bypass
            new_locations[
                location_id
            ].auto_bypass_low_battery = self.auto_bypass_low_battery

            # set the usercode for the location
            usercode = (
                self.usercodes.get(location_id)
                or self.usercodes.get(str(location_id))
                or self.usercodes.get("default")
            )
            if usercode:
                new_locations[location_id].usercode = usercode
            else:
                LOGGER.warning(f"no usercode for location {location_id}")
                new_locations[location_id].usercode = DEFAULT_USERCODE

        self.times["_make_locations()"] = time.time() - start_time
        return new_locations

    def arm_away(self, location_id):
        """Arm the system (Away)."""
        warnings.warn(
            "Using deprecated client.arm_away(). " "Use location.arm_away().",
            DeprecationWarning,
        )
        return self.arm(ArmType.AWAY, location_id)

    def arm_stay(self, location_id):
        """Arm the system (Stay)."""
        warnings.warn(
            "Using deprecated client.arm_stay(). " "Use location.arm_stay().",
            DeprecationWarning,
        )
        return self.arm(ArmType.STAY, location_id)

    def arm_stay_instant(self, location_id):
        """Arm the system (Stay - Instant)."""
        warnings.warn(
            "Using deprecated client.arm_stay_instant(). "
            "Use location.arm_stay_instant().",
            DeprecationWarning,
        )
        return self.arm(ArmType.STAY_INSTANT, location_id)

    def arm_away_instant(self, location_id):
        """Arm the system (Away - Instant)."""
        warnings.warn(
            "Using deprecated client.arm_away_instant(). "
            "Use location.arm_away_instant().",
            DeprecationWarning,
        )
        return self.arm(ArmType.AWAY_INSTANT, location_id)

    def arm_stay_night(self, location_id):
        """Arm the system (Stay - Night)."""
        warnings.warn(
            "Using deprecated client.arm_stay_night(). "
            "Use location.arm_stay_night().",
            DeprecationWarning,
        )
        return self.arm(ArmType.STAY_NIGHT, location_id)

    def arm(self, arm_type, location_id):
        """Arm the system. Return True if successful."""
        warnings.warn(
            "Using deprecated client.arm(). " "Use location.arm().", DeprecationWarning
        )
        return self.locations[location_id].arm(arm_type)

    def get_custom_arm_settings(self, location_id):
        """Get custom arm settings.  Return true if successful."""
        warnings.warn(
            "Using deprecated client.get_custom_arm_settings(). "
            "Use location.get_custom_arm_settings().",
            DeprecationWarning,
        )
        return self.locations[location_id].get_custom_arm_settings()

    def get_panel_meta_data(self, location_id):
        """Get all meta data about the alarm panel."""
        warnings.warn(
            "Using deprecated client.get_panel_meta_data(). "
            "Use location.get_panel_meta_data().",
            DeprecationWarning,
        )
        return self.locations[location_id].get_panel_meta_data()

    def zone_status(self, location_id, zone_id):
        """Get status of a zone."""
        warnings.warn(
            "Using deprecated client.zone_status(). " "Use location.zone_status().",
            DeprecationWarning,
        )
        return self.locations[location_id].zone_status(zone_id)

    def disarm(self, location_id):
        """Disarm the system. Return True if successful."""
        warnings.warn(
            "Using deprecated client.disarm(). " "Use location.disarm().",
            DeprecationWarning,
        )
        return self.locations[location_id].disarm()

    def zone_bypass(self, zone_id, location_id):
        """Bypass a zone.  Return true if successful."""
        warnings.warn(
            "Using deprecated client.zone_bypass(). " "Use location.zone_bypass().",
            DeprecationWarning,
        )
        return self.locations[location_id].zone_bypass(zone_id)

    def get_zone_details(self, location_id):
        """Get Zone details. Return True if successful."""
        warnings.warn(
            "Using deprecated client.get_zone_details(). "
            "Use location.get_zone_details().",
            DeprecationWarning,
        )
        return self.locations[location_id].get_zone_details()


class ArmingHelper:
    """
    For a partition or location, you can call its arm() or disarm() method directly.
       Example: partition.arm(ArmType.AWAY)

    Alternatively, you can use ArmingHelper.
       Example: ArmingHelper(partition).arm_away()
    """

    def __init__(self, partition_or_location):
        self.armable = partition_or_location

    def arm_away(self):
        """Arm the system (Away)."""
        self.armable.arm(ArmType.AWAY)

    def arm_stay(self):
        """Arm the system (Stay)."""
        self.armable.arm(ArmType.STAY)

    def arm_stay_instant(self):
        """Arm the system (Stay - Instant)."""
        self.armable.arm(ArmType.STAY_INSTANT)

    def arm_away_instant(self):
        """Arm the system (Away - Instant)."""
        self.armable.arm(ArmType.AWAY_INSTANT)

    def arm_stay_night(self):
        """Arm the system (Stay - Night)."""
        self.armable.arm(ArmType.STAY_NIGHT)

    def disarm(self):
        self.armable.disarm()



"""Total Connect Partition."""

import logging

from .const import ArmingState
from .exceptions import PartialResponseError


LOGGER = logging.getLogger(__name__)


class TotalConnectPartition:
    """Partition class for Total Connect."""

    def __init__(self, details, parent):
        """Initialize Partition based on PartitionDetails."""
        self.parent = parent
        self.id = details.get("PartitionID")
        self.name = details.get("PartitionName")
        self.update(details)

    def __str__(self):
        """Return a string that is printable."""
        data = f"PARTITION {self.id} - {self.name}\n" f"  {self.arming_state}\n"
        return data

    def update(self, info):
        """Update partition based on PartitionInfo."""
        astate = (info or {}).get("ArmingState")
        if astate is None:
            raise PartialResponseError("no ArmingState")
        try:
            self.arming_state = ArmingState(astate)
        except ValueError:
            LOGGER.error(f"unknown ArmingState {astate} in {result} -- please file an issue at https://github.com/craigjmidwinter/total-connect-client/issues")
            raise

    def arm(self, arm_type):
        """Arm the partition."""
        self.parent.arm(arm_type, self.id)

    def disarm(self):
        """Disarm the partition."""
        self.parent.disarm(self.id)



"""Total Connect Location."""

import logging

from .const import ArmingState, ArmType
from .device import TotalConnectDevice
from .exceptions import PartialResponseError, TotalConnectError
from .partition import TotalConnectPartition
from .zone import TotalConnectZone

DEFAULT_USERCODE = "-1"

LOGGER = logging.getLogger(__name__)


class TotalConnectLocation:
    """TotalConnectLocation class."""

    def __init__(self, location_info_basic, parent):
        """Initialize based on a 'LocationInfoBasic'."""
        self.location_id = location_info_basic["LocationID"]
        self.location_name = location_info_basic["LocationName"]
        self._photo_url = location_info_basic["PhotoURL"]
        self._module_flags = dict(
            x.split("=") for x in location_info_basic["LocationModuleFlags"].split(",")
        )
        self.security_device_id = location_info_basic["SecurityDeviceID"]
        self.parent = parent
        self.ac_loss = None
        self.low_battery = None
        self.cover_tampered = None
        self.last_updated_timestamp_ticks = None
        self.configuration_sequence_number = None
        self.arming_state = None
        self.partitions = {}
        self._partition_list = None
        self.zones = {}
        self.usercode = DEFAULT_USERCODE
        self._auto_bypass_low_battery = False

        dib = (location_info_basic.get("DeviceList") or {}).get("DeviceInfoBasic")
        tcdevs = [TotalConnectDevice(d) for d in (dib or {})]
        self.devices = {tcdev.id: tcdev for tcdev in tcdevs}

    def __str__(self):
        """Return a text string that is printable."""
        data = (
            f"LOCATION {self.location_id} - {self.location_name}\n\n"
            f"PhotoURL: {self._photo_url}\n"
            f"SecurityDeviceID: {self.security_device_id}\n"
            f"AcLoss: {self.ac_loss}\n"
            f"LowBattery: {self.low_battery}\n"
            f"IsCoverTampered: {self.cover_tampered}\n"
            f"{self.arming_state}\n"
            f"LocationModuleFlags:\n"
        )

        for key, value in self._module_flags.items():
            data = data + f"  {key}: {value}\n"

        data = data + "\n"

        devices = f"DEVICES: {len(self.devices)}\n\n"
        for device in self.devices:
            devices = devices + str(self.devices[device]) + "\n"

        partitions = f"PARTITIONS: {len(self.partitions)}\n\n"
        for partition in self.partitions:
            partitions = partitions + str(self.partitions[partition]) + "\n"

        zones = f"ZONES: {len(self.zones)}\n\n"
        for zone in self.zones:
            zones = zones + str(self.zones[zone])

        return data + devices + partitions + zones

    @property
    def auto_bypass_low_battery(self):
        """Return true if set to automatically bypass a low battery."""
        return self._auto_bypass_low_battery

    @auto_bypass_low_battery.setter
    def auto_bypass_low_battery(self, value: bool):
        """Set to automatically bypass a low battery."""
        self._auto_bypass_low_battery = value

    def set_zone_details(self, result):
        """
        Update from GetZonesListInStateEx_V1.

        ZoneStatusInfoWithPartitionId provides additional info for setting up zones.
        If we used TotalConnectZone.update() it would overwrite missing data with None.
        """
        zone_info = ((result.get("ZoneStatus") or {}).get("Zones") or {}).get(
            "ZoneStatusInfoWithPartitionId"
        )
        if not zone_info:
            raise PartialResponseError("no ZoneStatusInfoWithPartitionId", result)

        for zonedata in zone_info:
            self.zones[zonedata["ZoneID"]] = TotalConnectZone(zonedata)

    def get_panel_meta_data(self):
        """Get all meta data about the alarm panel."""
        # see https://rs.alarmnet.com/TC21api/tc2.asmx?op=GetPanelMetaDataAndFullStatus
        result = self.parent.request(
            f"GetPanelMetaDataAndFullStatusEx_V2(self.token, {self.location_id}, 0, 0, {self._partition_list})"
        )
        self.parent.raise_for_resultcode(result)

        self.set_status(result)
        self.update_partitions(result)
        self.update_zones(result)

    def set_status(self, result):
        """Update from result."""
        data = (result or {}).get("PanelMetadataAndStatus")
        if not data:
            raise PartialResponseError("no PanelMetadataAndStatus", result)

        self.ac_loss = data.get("IsInACLoss")
        self.low_battery = data.get("IsInLowBattery")
        self.cover_tampered = data.get("IsCoverTampered")
        self.last_updated_timestamp_ticks = data.get("LastUpdatedTimestampTicks")
        self.configuration_sequence_number = data.get("ConfigurationSequenceNumber")

        astate = result.get("ArmingState")
        if not astate:
            raise PartialResponseError("no ArmingState", result)
        try:
            self.arming_state = ArmingState(astate)
        except ValueError:
            LOGGER.error(f"unknown ArmingState {astate} in {result} -- please file an issue at https://github.com/craigjmidwinter/total-connect-client/issues")
            raise

    def get_zone_details(self):
        """Get Zone details."""
        result = self.parent.request(
            f"GetZonesListInStateEx_V1(self.token, {self.location_id}, {self._partition_list}, 0)"
        )

        if result["ResultCode"] == self.parent.FEATURE_NOT_SUPPORTED:
            LOGGER.warning(
                "getting Zone Details is a feature not supported by "
                "your Total Connect account or hardware"
            )
        self.parent.raise_for_resultcode(result)
        self.set_zone_details(result)

    def update_partitions(self, result):
        """Update partition info from Partitions."""
        pi = ((result.get("PanelMetadataAndStatus") or {}).get("Partitions") or {}).get(
            "PartitionInfo"
        )
        if not pi:
            raise PartialResponseError("no PartitionInfo", result)

        # loop through partitions and update
        # NOTE: do not use keys because they don't line up with PartitionID
        for partition in pi:
            if "PartitionID" not in partition:
                raise PartialResponseError("no PartitionID", result)
            partition_id = int(partition["PartitionID"])
            if partition_id in self.partitions:
                self.partitions[partition_id].update(partition)
            else:
                LOGGER.warning(f"Update provided for unknown partion {partition_id} ")

    def update_zones(self, result):
        """Update zone info from ZoneInfo or ZoneInfoEx."""

        data = (result.get("PanelMetadataAndStatus") or {}).get("Zones")
        if not data:
            LOGGER.error(
                f"no zones found: sync your panel using TotalConnect app or website"
            )
            # PartialResponseError would mean this is retryable without fixing
            # anything, and this needs fixing
            raise TotalConnectError("no zones found: panel sync required")

        zone_info = data.get("ZoneInfoEx") or data.get("ZoneInfo")
        if not zone_info:
            raise PartialResponseError("no ZoneInfoEx or ZoneInfo", result)
        for zonedata in zone_info:
            zid = (zonedata or {}).get("ZoneID")
            if not zid:
                raise PartialResponseError("no ZoneID", result)
            zone = self.zones.get(zid)
            if zone:
                zone.update(zonedata)
            else:
                zone = TotalConnectZone(zonedata)
                self.zones[zid] = zone

            if zone.is_low_battery() and self.auto_bypass_low_battery:
                self.parent.zone_bypass(zone, self.location_id)

    def get_partition_details(self):
        """Get partition details for this location."""
        # see https://rs.alarmnet.com/TC21api/tc2.asmx?op=GetPartitionsDetails

        result = self.parent.request(
            f"GetPartitionsDetails(self.token, {self.location_id}, {self.security_device_id})"
        )
        try:
            self.parent.raise_for_resultcode(result)
        except TotalConnectError:
            LOGGER.error(
                f"Could not get partition details for "
                f"device {self.security_device_id} at "
                f"location {self.location_id}."
                f"ResultCode: {result['ResultCode']}. "
                f"ResultData: {result['ResultData']}"
            )
            raise

        partition_details = ((result or {}).get("PartitionsInfoList") or {}).get(
            "PartitionDetails"
        )
        if not partition_details:
            raise PartialResponseError("no PartitionDetails", result)

        new_partition_list = []
        for partition in partition_details:
            new_partition = TotalConnectPartition(partition, self)
            self.partitions[new_partition.id] = new_partition
            new_partition_list.append(new_partition.id)

        self._partition_list = {"int": new_partition_list}

    def is_low_battery(self):
        """Return true if low battery."""
        return self.low_battery is True

    def is_ac_loss(self):
        """Return true if AC loss."""
        return self.ac_loss is True

    def is_cover_tampered(self):
        """Return true if cover is tampered."""
        return self.cover_tampered is True

    def set_usercode(self, usercode):
        """Set the usercode. Return true if successful."""
        if self.parent.validate_usercode(self.security_device_id, usercode):
            self.usercode = usercode
            return True
        return False

    def arm(self, arm_type, partition_id=None):
        """Arm the given partition. If no partition is given, arm all partitions."""
        # see https://rs.alarmnet.com/TC21api/tc2.asmx?op=ArmSecuritySystemPartitionsV1
        assert isinstance(arm_type, ArmType)
        partition_list = []
        if partition_id is None:
            partition_list = self._partition_list
        else:
            if partition_id not in self.partitions:
                raise TotalConnectError(
                    f"Partition {partition_id} does not exist "
                    f"at location {self.location_id}"
                )
            partition_list.append(partition_id)

        result = self.parent.request(
            f"ArmSecuritySystemPartitionsV1(self.token, "
            f"{self.location_id}, "
            f"{self.security_device_id}, "
            f"{arm_type.value}, "
            f"'{self.usercode}', "
            f"{partition_list})"
        )
        if result["ResultCode"] == self.parent.COMMAND_FAILED:
            LOGGER.warning("could not arm system; is a zone faulted?")
        self.parent.raise_for_resultcode(result)
        LOGGER.info(
            f"ARMED({arm_type}) partitions {partition_list} at {self.location_id}"
        )

    def disarm(self, partition_id=None):
        """Disarm the system."""
        # if no partition is given, disarm all partitions
        # see https://rs.alarmnet.com/TC21api/tc2.asmx?op=ArmSecuritySystemPartitionsV1
        partition_list = []
        if partition_id is None:
            partition_list = self._partition_list
        else:
            if partition_id not in self.partitions:
                raise TotalConnectError(
                    f"Partition {partition_id} does not exist "
                    f"at location {self.location_id}"
                )
            partition_list.append(partition_id)

        result = self.parent.request(
            f"DisarmSecuritySystemPartitionsV1(self.token, "
            f"{self.location_id}, "
            f"{self.security_device_id}, "
            f"'{self.usercode}', "
            f"{partition_list})"
        )
        self.parent.raise_for_resultcode(result)
        LOGGER.info(f"DISARMED partitions {partition_list} at {self.location_id}")

    def zone_bypass(self, zone_id):
        """Bypass a zone."""
        result = self.parent.request(
            f"Bypass(self.token, "
            f"{self.location_id}, "
            f"{self.security_device_id}, "
            f"{zone_id}, "
            f"'{self.usercode}')"
        )
        self.parent.raise_for_resultcode(result)
        LOGGER.info(f"BYPASSED {zone_id} at {self.location_id}")
        self.zones[zone_id]._mark_as_bypassed()

    def zone_status(self, zone_id):
        """Get status of a zone."""
        z = self.zones.get(zone_id)
        if not z:
            raise TotalConnectError(f"zone {zone_id} does not exist")
        return z.status

    def arm_custom(self, arm_type):
        """NOT OPERATIONAL YET.
        Arm custom the system.  Return true if successful.
        """
        ZONE_INFO = {"ZoneID": "12", "ByPass": False, "ZoneStatus": 0}

        ZONES_LIST = {}
        ZONES_LIST[0] = ZONE_INFO

        CUSTOM_ARM_SETTINGS = {"ArmMode": "1", "ArmDelay": "5", "ZonesList": ZONES_LIST}

        result = self.request(
            f"CustomArmSecuritySystem(self.token, "
            f"{self.location_id}, "
            f"{self.security_device_id}, "
            f"{arm_type.value}, '{self.usercode}', "
            f"{CUSTOM_ARM_SETTINGS})"
        )

        if result["ResultCode"] != self.SUCCESS:
            LOGGER.error(
                f"Could not arm custom. ResultCode: {result['ResultCode']}. "
                f"ResultData: {result['ResultData']}"
            )
            return False

        # remove after this is all working
        print(
            f"arm_custom ResultCode: {result['ResultCode']}. "
            f"arm_custom ResultData: {result['ResultData']}"
        )

        return True

    def get_custom_arm_settings(self):
        """NOT OPERATIONAL YET.
        Get custom arm settings.
        """
        result = self.parent.request(
            f"GetCustomArmSettings(self.token, "
            f"{self.location_id}, "
            f"{self.security_device_id})"
        )
        self.parent.raise_for_resultcode(result)
        # FIXME: returning the raw result is not right
        return result
